# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================


from twitter.common.lang import Compatibility, Singleton
from .gauge import (
  Gauge,
  MutatorGauge,
  NamedGauge,
  namablegauge)


class Observable(object):
  """
    A trait providing a metric namespace for an object.

    Classes should mix-in Observable and register metrics against self.metrics.

    Application owners can then register observable objects into a metric
    space or the root metrics, e.g. via

    >>> RootMetrics().register_observable('object_namespace', my_object)

  """
  @property
  def metrics(self):
    """
      Returns a Metric namespace for this object.
    """
    if not hasattr(self, '_observable_metrics'):
      self._observable_metrics = Metrics()
    return self._observable_metrics


class MetricProvider(object):
  def sample(self):
    """
      Returns a dictionary
        string (metric) => sample (number)
    """
    raise NotImplementedError


class MetricRegistry(object):
  def scope(self, name):
    """
      Returns a (potentially memoized) child scope with a given name.
    """
    raise NotImplementedError

  def register(self, gauge):
    """
      Register a gauge (mapper from name => sample) with this registry.
    """
    raise NotImplementedError

  def unregister(self, name):
    """
      Unregister a name from the registry.
    """
    raise NotImplementedError

  def mutator(self, name):
    """
      Return a mutator function of the gauge associated with name.
    """
    raise NotImplementedError


class Metrics(MetricRegistry, MetricProvider):
  """
    Metric collector.
  """

  class Error(Exception): pass

  @classmethod
  def coerce_value(cls, value):
    if isinstance(value, Compatibility.numeric + Compatibility.string + (bool,)):
      return value
    elif value is None:
      return value
    elif isinstance(value, list):
      return [cls.coerce_value(v) for v in value]
    elif isinstance(value, dict):
      return dict((cls.coerce_value(k), cls.coerce_value(v)) for (k, v) in value.items())
    else:
      return str(value)

  @classmethod
  def coerce_metric(cls, metric_tuple):
    name, value = metric_tuple
    try:
      return (name, cls.coerce_value(value.read()))
    except ValueError:
      return None

  def __init__(self):
    self._metrics = {}
    self._children = {}

  def scope(self, name):
    if not isinstance(name, Compatibility.string):
      raise TypeError('Scope names must be strings, got: %s' % type(name))
    if name not in self._children:
      self._children[name] = Metrics()
    return self._children[name]

  def register_observable(self, name, observable):
    if not isinstance(name, Compatibility.string):
      raise TypeError('Scope names must be strings, got: %s' % type(name))
    if not isinstance(observable, Observable):
      raise TypeError('observable must be an Observable, got: %s' % type(observable))
    self._children[name] = observable.metrics

  def unregister_observable(self, name):
    if not isinstance(name, Compatibility.string):
      raise TypeError('Unregister takes a string name!')
    return self._children.pop(name, None)

  def register(self, gauge):
    if isinstance(gauge, Compatibility.string):
      gauge = MutatorGauge(gauge)
    if not isinstance(gauge, NamedGauge) and not namablegauge(gauge):
      raise TypeError('Must register either a string or a Gauge-like object! Got %s' % gauge)
    self._metrics[gauge.name()] = gauge
    return gauge

  def unregister(self, name):
    if not isinstance(name, Compatibility.string):
      raise TypeError('Unregister takes a string name!')
    return self._metrics.pop(name, None)

  @classmethod
  def sample_name(cls, scope_name, sample_name):
    return '.'.join([scope_name, sample_name])

  def sample(self):
    samples = dict(filter(None, map(self.coerce_metric, self._metrics.items())))
    for scope_name, scope in self._children.items():
      samples.update((self.sample_name(scope_name, sample_name), sample_value)
                     for (sample_name, sample_value) in scope.sample().items())
    return samples


class CompoundMetrics(MetricProvider):
  def __init__(self, *providers):
    if not all(isinstance(provider, MetricProvider) for provider in providers):
      raise TypeError('CompoundMetrics must take a collection of MetricProviders')
    self._providers = providers

  def sample(self):
    root_sample = {}
    for provider in self._providers:
      root_sample.update(provider.sample())
    return root_sample


class MemoizedMetrics(MetricProvider):
  def __init__(self, provider):
    if not isinstance(provider, MetricProvider):
      raise TypeError('MemoizedMetrics must take a MetricProvider')
    self._provider = provider
    self._sample = {}

  def sample(self):
    self._sample = self._provider.sample()
    return self._sample

  @property
  def memoized_sample(self):
    return self._sample


class RootMetrics(Metrics, Singleton):
  """
    Root singleton instance of the metrics.
  """

  _INIT = False

  def __init__(self):
    if not RootMetrics._INIT:
      Metrics.__init__(self)
      RootMetrics._INIT = True

  # For testing.
  def clear(self):
    Metrics.__init__(self)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import time

from twitter.common.quantity import Amount, Time
from .gauge import NamedGauge, gaugelike, namablegauge

class Rate(NamedGauge):
  """
    Gauge that computes a windowed rate.
  """
  @staticmethod
  def of(gauge, name = None, window = None, clock = None):
    kw = {}
    if window: kw.update(window = window)
    if clock: kw.update(clock = clock)
    if name:
      if not gaugelike(gauge):
        raise TypeError('Rate.of must take a Gauge-like object!  Got %s' % type(gauge))
      return Rate(name, gauge, **kw)
    else:
      if not namablegauge(gauge):
        raise TypeError('Rate.of must take a namable Gauge-like object if no name specified!')
      return Rate(gauge.name(), gauge, **kw)

  def __init__(self, name, gauge, window = Amount(1, Time.SECONDS), clock = time):
    """
      Create a gauge using name as a base for a <name>_per_<window> sampling gauge.

        name: The base name of the gauge.
        gauge: The gauge to sample
        window: The window over which the samples should be measured (default 1 second.)
    """
    self._clock = clock
    self._gauge = gauge
    self._samples = []
    self._window = window
    NamedGauge.__init__(self, '%s_per_%s%s' % (name, window.amount(), window.unit()))

  def filter(self, newer_than=None):
    """
      Filter the samples to only contain elements in the window.
    """
    if newer_than is None:
      newer_than = self._clock.time() - self._window.as_(Time.SECONDS)
    self._samples = [sample for sample in self._samples if sample[0] >= newer_than]

  def read(self):
    now = self._clock.time()
    self.filter(now - self._window.as_(Time.SECONDS))
    new_sample = self._gauge.read()
    self._samples.insert(0, (now, new_sample))
    if len(self._samples) == 1:
      return 0
    last_sample = self._samples[-1]
    dy = new_sample - last_sample[1]
    dt = now - last_sample[0]
    return 0 if dt == 0 else dy / dt



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import json
import os
import time
import threading

try:
  from twitter.common import log
except ImportError:
  log = None

from twitter.common.exceptions import ExceptionalThread
from twitter.common.quantity import Amount, Time

from .metrics import MetricProvider


class SamplerBase(ExceptionalThread):
  def __init__(self, period, clock):
    self._stop = threading.Event()
    self._period = period
    self._clock = clock
    ExceptionalThread.__init__(self)
    self.daemon = True

  def stop(self):
    self._stop.set()

  def is_stopped(self):
    return self._stop.is_set()

  def iterate(self):
    raise NotImplementedError

  def run(self):
    while True:
      self._clock.sleep(self._period.as_(Time.SECONDS))
      if self.is_stopped():
        break
      self.iterate()


class MetricSampler(SamplerBase, MetricProvider):
  """
    A thread that periodically samples from a MetricProvider and caches the
    samples.
  """
  def __init__(self, provider, period=Amount(1, Time.SECONDS), clock=time):
    self._provider = provider
    self._last_sample = self._provider.sample()
    self._lock = threading.Lock()
    SamplerBase.__init__(self, period, clock)
    self.daemon = True

  def sample(self):
    with self._lock:
      return self._last_sample

  def iterate(self):
    new_sample = self._provider.sample()
    with self._lock:
      self._last_sample = new_sample


class DiskMetricWriter(SamplerBase):
  """
    Takes a MetricProvider and periodically samples its values to disk in JSON format.
  """

  def __init__(self, provider, filename, period=Amount(15, Time.SECONDS), clock=time):
    self._provider = provider
    self._filename = filename
    SamplerBase.__init__(self, period, clock)
    self.daemon = True

  def iterate(self):
    with open(self._filename, 'w') as fp:
      json.dump(self._provider.sample(), fp)


class DiskMetricReader(SamplerBase, MetricProvider):
  """
    Given an input JSON file, periodically reads the contents from disk and exports it
    using the MetricProvider interface.
  """

  def __init__(self, filename, period=Amount(15, Time.SECONDS), clock=time):
    self._filename = filename
    self._sample = {}
    self._lock = threading.Lock()
    SamplerBase.__init__(self, period, clock)
    self.daemon = True

  def sample(self):
    with self._lock:
      return self._sample

  @property
  def age(self):
    try:
      return time.time() - os.path.getmtime(self._filename)
    except (IOError, OSError):
      return 0

  def iterate(self):
    with self._lock:
      try:
        with open(self._filename, 'r') as fp:
          self._sample = json.load(fp)
      except (IOError, OSError, ValueError) as e:
        if log:
          log.warn('Failed to collect sample: %s' % e)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================


from twitter.common.lang import Compatibility


# Duck-typing helpers
def gaugelike(obj):
  return hasattr(obj, 'read') and callable(obj.read)

def namable(obj):
  return hasattr(obj, 'name') and callable(obj.name)

def namablegauge(obj):
  return gaugelike(obj) and namable(obj)


# Typed gauges.
class Gauge(object):
  """
    A readable gauge that exports a value.
  """
  def __init__(self, value):
    self._value = value

  def read(self):
    return self._value


class NamedGauge(Gauge):
  """
    Named gauge (gauge that exports name() method.)
  """
  def __init__(self, name, value=None):
    if not isinstance(name, str):
      raise TypeError('NamedGauge must be named by a string, got %s' % type(name))
    self._name = name
    Gauge.__init__(self, value)

  def name(self):
    return self._name


class MutableGauge(Gauge):
  """
    Mutable gauge.
  """
  def __init__(self, value=None):
    import threading
    self._lock = threading.Lock()
    Gauge.__init__(self, value)

  def read(self):
    with self.lock():
      return self._value

  def write(self, value):
    with self.lock():
      self._value = value
      return self._value

  def lock(self):
    return self._lock


class Label(NamedGauge):
  """
    A generic immutable key-value Gauge.  (Not specifically strings, but that's
    the intention.)
  """
  def __init__(self, name, value):
    NamedGauge.__init__(self, name, value)


class LambdaGauge(NamedGauge):
  def __init__(self, name, fn):
    import threading
    if not callable(fn):
      raise TypeError("A LambdaGauge must be supplied with a callable, got %s" % type(fn))
    NamedGauge.__init__(self, name, fn)
    self._lock = threading.Lock()

  def read(self):
    with self._lock:
      return self._value()


class MutatorGauge(NamedGauge, MutableGauge):
  def __init__(self, name, value=None):
    NamedGauge.__init__(self, name)
    MutableGauge.__init__(self, value)


class AtomicGauge(NamedGauge, MutableGauge):
  """
    Something akin to AtomicLong. Basically a MutableGauge but with
    atomic add, increment, decrement.
  """
  def __init__(self, name, initial_value=0):
    if not isinstance(initial_value, Compatibility.integer):
      raise TypeError('AtomicGauge must be initialized with an integer.')
    NamedGauge.__init__(self, name)
    MutableGauge.__init__(self, initial_value)

  def add(self, delta):
    """
      Add delta to metric and return updated metric.
    """
    if not isinstance(delta, Compatibility.integer):
      raise TypeError('AtomicGauge.add must be called with an integer.')
    with self.lock():
      self._value += delta
      return self._value

  def increment(self):
    """
      Increment metric and return updated metric.
    """
    return self.add(1)

  def decrement(self):
    """
      Decrement metric and return updated metric.
    """
    return self.add(-1)
